package pinball;

import pinball.*;
import pinball.buffer.*; 
import pinball.part.*;
import pinball.server.*;
import pinball.player.*;

import java.io.*;

/**
 * The monitor for the locks on the threads. 
 */
public class Watcher implements Serializable{ 
	/** Used for serialization. */
	static final long serialVersionUID = -2322229476047190392L;


	/** The values for the lock positions. */	
	public final int empty = 0;

	/** Buffer frame was grabed. */
	public final int grabed = 1;

	/** Buffer frame was analyzed. */ 
	public final int analyzed = 3;

	/** Buffer frame was used to make the flipping descision. */
	public final int strategied = 4;

	/** Buffer frame was drawn on the screen. */
	public final int posted = 5;		
	
	/** Buffer is in frame grabber. */
	public final int busyGrabber =-1;
		 
	/** Buffer is in analysis. */
	public final int busyAnalysis =-3;

	/** Buffer is in strategy. */
	public final int busyStrategy =-4;

	/** Buffer is in post. */
	public final int busyPost =-5;

	/** Lock status. */		
        private int status=empty;

	/* Lock states for the next position. */
	
	/** Next lock state for grab thread. */
	private int grabNext=grabed;

	/** Next lock state for analyze thread. */
	private int analyzeNext=analyzed;
	
	/** Next lock state for strategy thread. */
	private int strategyNext=strategied;
	
	/** Next lock state for post thread. */
	private int postNext=empty;
	
	/** Who accessed the lock. */
	private int id;

	/**
	 *  Setup the lock for the threads. 
	 */
	public Watcher() {
		status=empty;
		
		/* Setup the positions of the lock depending what
		   threads are being used. 
		*/
		if(!GameOptions.post())
	   		strategyNext=postNext;
		if(!GameOptions.strategy())
	   		analyzeNext=strategyNext;
		if(!GameOptions.analyze())
			grabNext=analyzeNext;
	}

	/** 
	 * Start grab. Wait for frame to be empty.
	 *
	 * @param id id of buffer, used for debugging.
	 */
        public synchronized void grab(int d) {
               	id=d;
                while (status!=empty) {
        	        try {
                       	if(GameOptions.debug()>=2)
	                	System.err.println("*"+id+"*     "+"Can't Grab Frame: " + status);
                        	wait();
                        } 
                        catch (InterruptedException e) {System.err.println(e);}
                }
		if(GameOptions.debug()>=2)
                       	System.err.println("*"+id+"*     "+"Frame is Grabable.");
		status=busyGrabber;                    
	}	
		
	/** 
	 * End grab. Set frame to next stage status.
	 *
	 * @param id id of buffer, used for debugging.
	 */
	public synchronized void grabed(int d) {
 		id=d;
 		status=grabNext;
		notifyAll();
                if(GameOptions.debug()>=2)
		        System.err.println("*"+id+"*     "+"Frame Grabbed.");                 
	                     
	}

	/** 
	 * Start analyze. Wait for frame to be grabbed.
	 *
	 * @param id id of buffer, used for debugging.
	 */
        public synchronized void analyze(int d) {
		id=d;
                while (status!=grabed) {
			try {
				if(GameOptions.debug()>=2)
				System.err.println("*"+id+"*     "+"Image not ready for analysis: " + status);
                            	wait();
			} 
			catch (InterruptedException e) {System.err.println(e);}
		}
                if(GameOptions.debug()>=2)                                  	
			System.err.println("*"+id+"*     "+"Image is analyizible " + status);     
           
                        status=busyAnalysis;
		}

	/** 
	 * End analyze. Set frame to next stage status.
	 *
	 * @param id id of buffer, used for debugging.
	 */
 	public synchronized void analyzed(int d) {
		id=d;
		status=analyzeNext;  
               	notifyAll();
 		if(GameOptions.debug()>=2)
			System.err.println("*"+id+"*     "+"Frame Analyzation Complete.");          
				         
	}

		
	/**
	 * Start strategy. Wait for frame to be analyzed.
	 *
	 * @param id id of buffer, used for debugging.
	 */
	public synchronized void strategy(int d) {
		id=d;
                while (status!=analyzed) {
                	try {
				if(GameOptions.debug()>=2)
			        System.err.println("*"+id+"*     "+"Image not ready for Strategy: " + status);
        	                wait();
			} 
			catch (InterruptedException e) {System.err.println(e);}
		}
                if(GameOptions.debug()>=2)                                                   	    			
			System.err.println("*"+id+"*     "+"Image is ready for Strategy " + status);     
                        status=busyStrategy;
	}

	/** 
	 * End strategy. Set frame to next stage status.
	 *
	 * @param id id of buffer, used for debugging.
	 */
	public synchronized void strategied(int d) {
		id=d;
		status=strategyNext;  
            	notifyAll();
 		if(GameOptions.debug()>=2)
			System.err.println("*"+id+"*     "+"Strategy Complete.");          				         
        }


	/**
	 * Start display. Wait for frame to got through strategy.
	 *
	 * @param id id of buffer, used for debugging.
	 */
        public synchronized void post(int d) {
		id=d;
                while (status!=strategied) {
                	try {
                        	if(GameOptions.debug()>=2)
	                         System.err.println("*"+id+"*     "+"Image not postable: " + status);
	                            wait();
                        } 
                        catch (InterruptedException e) {System.err.println(e);}
                }
		if(GameOptions.debug()>=2)
			System.err.println("*"+id+"*     "+"Image is postable " + status);                 
                        status=busyPost;
	}
		
	/** 
	 * End display. Set frame to next stage status.
	 *
	 * @param id id of buffer, used for debugging.
	 */
	public synchronized void posted(int d) {
		id=d;
		status=postNext;
               	notifyAll();
	    	if(GameOptions.debug()>=2)	
			System.err.println("*"+id+"*     "+"Frame Done With.");              
	}
               
	/** 
	 * Find the status of the lock. 
	 *
	 * @return the status of the lock.
	 */
	public synchronized int status() {
	      		return status;
	}

}
