package pinball.buffer;

import pinball.*;
import pinball.buffer.*;
import pinball.io.*;
import pinball.part.*;
import pinball.server.*;
import pinball.player.*;

import java.io.*;
import java.util.*;

/** 
 * The information buffer's information.
 * This contains data about the frame from the camera.
 * It is passed around to each of the threads. Each thread makes whatever
 * analysis it needs to do on the information so that the system can play
 * pinball. 
 * @see pinball.buffer.Buffer
 */
public class BufferInfo implements Serializable{
	/** Used to keep the same serialized version. */
	static final long serialVersionUID = -6577713727559496460L;

	/** The frame from the camera. */
	protected  CameraFrame frame=null;
	
	/** The frame of found objects. */
	protected  int [][] blips;
	
	/** Table of found objects. */
	protected  Vector objects=new Vector();

	/** The id of the information buffer. */
	protected int id;
	
	/** The number of the frame grabed. */
	protected int number;
	
	/** The time stamper, used to keep track of the time used by various processes. */
	protected transient TimeStamp timestamp;
	
	/** The object found. */
	protected Blip blip;
	
	/** The size of the frame. */
	protected int width,height;
	
	/** The x & y offsets for the image. */	
	protected int XOffset,YOffset;
	
	/** Are the frames to be cleared on the window. */
	protected static boolean clearing=false; 

	/** The y offset for the flippers. */
	protected int flipperOffset=0;
	
	/** Did the flipper flip. */
	protected boolean flipperFlipped=false;

	/** The image offset. */
	protected static int imageOffset=0;
	
	/** Number of bytes for the image. */
	protected int size;
	
	/** Used to define the top & bottom of the image to see if the ball is in play or out of 
	   play. 
	*/
	protected static int top=-1, bottom=-1;
	
	/** The resize information for the image. */
	protected static ResizeInfo resizeInfo=new ResizeInfo();
	
	/** The data IO device. */	
	protected static DataIO dataIO;
	
	/** The stub for the strategy. */
	protected static Stub stub;	
	
	
	/** 
	 *Build the buffer information. 
	 *
	 * @param id id number for buffer.
	 */
	public BufferInfo(int id) {	
		this.id=id;
		number=0;
		blip=new Blip();
		timestamp=new TimeStamp();

		if(dataIO==null) {
			dataIO=new DataIO(GameOptions.dataPort());
			GameOptions.setDataIO(dataIO);
		}

		
	}
	
	/**
	 * For extended class.
	 */
	public BufferInfo() {}

	/** 
	 *Get the buffer's id. 
	 *
	 * @return the buffer id.
	 */
	public int id() {
		return id;
	}
	
	
	/** 
	 * Make the frame for the image. 
	 *
	 * @param size the frame.
 	 * @param width of the frame.
  	 * @param height the frame.
	 */
	public void cameraInfo(int size, int width, int height) {
		this.size=size;
		this.width=width;
		this.height=height;
		
		frame=new CameraFrame(number,size,width,height,XOffset,YOffset);
	}

	/** 
	 * Make the frame for the image. 
	 *
	 * @param size the frame.
	 */
	public void cameraInfo(int size) {
		frame=new CameraFrame(number,size,width,height,XOffset,YOffset);
	}

	
	/**
	 * Get the frame, used to pass buffer to camera. 
	 *
	 * @return frame to store image.
	 */
	public byte [] get() {
		if(frame==null)
			return null;
		else return  frame.getFrame();
	}

	/**
	 * Get the CameraFrame for saving in a log file. 
	 *
	 * @return frame from camera.
	 */
	public CameraFrame getCameraFrame() {
		return frame;
	}

	/** 
	 * Set the frame buffer. 
	 *
	 * @param b frame to write to.
	 */
	 public void set(CameraFrame f) {
		frame=f;
		size=f.getFrame().length;
		width=f.getWidth();
		height=f.getHeight();
		XOffset=f.getXOffset();
		YOffset=f.getYOffset();
	}
	

	/** 
	 * Store the found objects frame and Vector table. 
	 *
	 * @param b the blips frame.
	 * @param f The objects found.
	 */
	public void setBlips(int [][] b, Vector f){
		blips=b;
		objects=f;
	}
	

	/** 
	 * Store the objects found. 
	 *
	 * @param f The objects found.
	 */
	public void setBlips(Vector f){
		objects=f;
	}
	
	/**
	 * Get the frame of found objects. 
	 *
	 * @return the blips frame.
	 */
	public int [][] getBlipsFrame() {
		return blips;
	}
	
	/**
	 * Get the objects table. 
	 */
	public Vector getObjects() {
		return objects;
	}
	
	/** 
	 * Set frame grabbed number. 
	 *
	 * @param number the number of the frame.
	 */
	public void number(int number) {
		this.number=number;
	}

	/**
	 * Get frame grabbed number. 
	 *
	 * @return the number of this frame.
	 */	
	public int number() {
		return number;
	}
		

	/** 
	 * Get the blip (ball) seen.
	 *
	 * @return blip.
	 */
	public Blip blip() {
		return blip;
	}


	/** 
	 * The time stamp used in the threads. 
	 *
	 * @return the time stamp.
	 */
	public TimeStamp time() {
		return timestamp;
	}


	/** 
	 * Set frame width. 
	 *
	 * @param width of frame.
	 */       
	public void width(int w) {
		width=w;
	}

	/** 
	 * Set frame height. 
	 *
	 * @param height of frame.
	 */       
	public void height(int h) {
		height=h;
	}
	
	/**
	 * Get the width of the frame. 
	 *
	 * @return width.
	 */
	public int width() {
		return width;
	}
	
	/**
	 * Get the height. 
	 *
	 * @return height.
	 */
	public int height() {
		return height;
	}	

	
	/**
	 * Set the location of the offsets.
	 *
	 * @param x xoffset.
	 * @param y yoffset.
	 */
	public void offsets(int x, int y) {
		XOffset=x;
		YOffset=y;
	}	
	
	/**
	 * Get the resize x size.
	 *
	 * @return x resize.
	 */
	public int resizeX() {
		return resizeInfo.width();
	}

	/**
	 * Get the resize y size.
	 *
	 * @return y resize.
	 */	
	public int resizeY() {
		return resizeInfo.height();
	}

	/**
	 * Get the resize x offset.
	 *
	 * @return x resize offset.
	 */
	public int resizeXOffset() {
		return resizeInfo.xoffset();
	}

	/**
	 * Get the resize y offset.
	 *
	 * @return y resize offset.
	 */
	public int resizeYOffset() {
		return resizeInfo.yoffset();
	}


	/**
	 * Get the x offset.
	 *
	 * @return x offset.
	 */
	public int XOffset() {
		return XOffset;
	}
	
	/**
	 * Get the y offset.
	 *
	 * @return y offset.
	 */
	public int YOffset() {
		return YOffset;
	}


	/** 
	 * Set the resize flag and include resize information. 
	 *
	 * @param x x size.
	 * @param y y size.
	 * @param OX x offset.
	 * @param OY y offset.
	 */
	public void resize(int x, int y, int OX, int OY) {
		resizeInfo.setSize(x,y,OX,OY);
	}
	
	/**
	 * Get the resize flag. 
	 * 
	 * @return resize flag.
	 */
	public boolean resize() {
		return resizeInfo.resize();
	}

	/**
	 * Set the condition that it was resized. 
	 */
	public void resized() {	
		resizeInfo.resized();		
	}
	
	/** 
	 * Set the clearing flag. Will clear the frame on the display. Used
	 * to clear when the ball's location is draw near the edge of the display.
	 * Clears the cross hair that's left at the edge.
	 *
	 * @param c clearing status.
	 */
	public void clearing(boolean c) {
		clearing=c;
	}

	/** 
	 * Get the clearing flag. 
	 *
	 * @param clearing status.
	 */
	public boolean clearing() {
		return clearing;
	}

	/**
	 * Get the flipper offset. 
	 *
	 * Return flipper offset.
	 */
	public int flipperOffset() {
		return this.flipperOffset;
	}
	
	/**
	 * Set the flipper offset. 
	 *
	 * @param i flipper offset.
	 */
	public void flipperOffset(int i) {
		this.flipperOffset=i;
	}
	
	/**
	 * Set the image offset. 
	 *
	 * @param i the image offset.
	 */
	public void imageOffset(int i) {
		imageOffset=i;
	}
	
	/** 
	 * Get the image offset. 
	 *
	 * @return the image offset.
	 */
	public int imageOffset() {
		return imageOffset;
	}
	
	/**
	 * Set the flippers flipped condition. 
	 *
	 * @param status the flippers flipped status.
	 */
	public void flipperFlipped(boolean status) {
		flipperFlipped=status;
	}
	
	/**
	 * Get the flippers flipped condition. 
	 *
	 * @return the status of the flippers flipped status.
	 */
	public boolean flipperFlipped() {
		return flipperFlipped;
	}

	/**
	 * Set the top & bottom for the in play area. 
  	 * Top is the top of the screen to define anything over that is to be
  	 * in play.
  	 * Bottom is the end of the playing area. Anything under that is out of
  	 * play.
  	 * 
  	 * @param t top of playfield.
  	 * @param b bottom of playfield.
  	 */
	public void topBottom(int t, int b) {
		top=t;
		bottom=b;
	}
	
	/**
	 * The top of the playfield.
	 *
	 * @return top of playfield.
	 */
	public int top() {
		return top;
	}

	/**
	 * The bottom of the playfield.
	 *
	 * @return bottom of playfield.
	 */	
	public int bottom() {
		return bottom;
	}

	/**
	 * Set the stub for the system.
	 *
	 * @param s The stub. 
	 */
	public void setStub(Stub s) {
		stub=s;
	}

	/**
	 * Get the stub for the system.
	 *
	 * @return The stub. 
	 */
	public Stub getStub() {
		return stub;
	}
}

