package pinball.io;

import java.io.*;
import java.util.*;
import pinball.*;

/** 
 * Controls for the RS232SDD16 data IO device. Currently the LSB is the output
 * part of the I/O unit sending data to the pinball machine. The MSB is used
 * to read the data from the I/O device.
 */
public class RS232SDD16 {
	/** Size of word for the device. */
	public final int WORDSIZE = 8;
	
	/** The serial port to write the data to. */
	private SerialIO serialIO;

	/** A zero bitset. */
	public BitSet ZERO;	

	/** A bitset with all the bits on. */
	public BitSet ALL;	
	
	/** The most significant bit set. */
	private BitSet msbSet;
	
	/** The least significant bit set. */	
	private BitSet lsbSet;
	
	/** The test port. */
	private static final String testPort = "/dev/cuaa0";
	
	/** Is the device connected. 
	Set this to false, if you don't have this device, 
	but want to run the draw player. */
	private boolean deviceConnected=true;
		
	
	/**
	 * Create the IO device at the port specified. 
	 *
	 * param port serial port location.
	 */
	public RS232SDD16 (String port) {
		if(GameOptions.client())
			return;
		serialIO=new SerialIO(port);
		
		msbSet=new BitSet(WORDSIZE);
		lsbSet=new BitSet(WORDSIZE);
		
		ZERO=new BitSet(WORDSIZE);
		
		ALL=new BitSet(WORDSIZE);
		
		/* Turn on all the bits for the ALL bitset. */
		for(int i=0;i<ALL.size();i++) {
			ALL.set(i);
		}


		/* Configure the device. */
		try{
			defineLines((byte)0,(byte)255);
			setPowerupStates((byte)0,(byte)0);			
		}
		
		catch(Exception e){
			System.err.println("Problem setting up the RS232SDD16");
			System.exit(1);
		}
		
		System.err.println("RS232SDD16 made.");
		
	}
	
 	/**
 	 * Clear the output lines on the device.
	 *
	 * param msb the most significant bit.
	 * param lsb the least significant bit.
	 */
	public synchronized void resetOutputLines(BitSet msb, BitSet lsb) throws IOException{		
		/* Turn off the bits. */
		for(int i=0;i<msb.size();i++) {
			if(msb.get(i))
				msbSet.clear(i);
		}

		/* Turn off the bits. */		
		for(int i=0;i<lsb.size();i++) {
			if(lsb.get(i))
				lsbSet.clear(i);
		}
		
		setOutputLines(msbSet,lsbSet);
	}

	/**
 	 * Set the output lines on the device.
	 *
	 * param msb the most significant bit.
	 * param lsb the least significant bit.
	 */
	public synchronized void setOutputLines(BitSet msb, BitSet lsb) throws IOException{		
		int mb=0, lb=0;

		/* Turn on the bits. */
		for(int i=0;i<msb.size();i++) {
			if(msb.get(i)) 
				msbSet.set(i);				
			if(msbSet.get(i))
				mb+=(int)Math.pow(2,i);
		}
		
		/* Turn on the bits. */
		for(int i=0;i<lsb.size();i++) {
			if(lsb.get(i))
				lsbSet.set(i);
			if(lsbSet.get(i))
				lb+=(int)Math.pow(2,i);
			
		}
		
		serialIO.write("!0SO"+(char)mb+(char)lb);
		

		/* Wait until the commands are sent to the lsb of the data unit. */
		//requestRead();
		//byte data[]=new byte[2];		
		//requestRead();
		//while(data[1]!=lb) {
		//	/* Send the command again. */
		//	serialIO.write("!0SO"+(char)mb+(char)lb);
		//	try{
		//		data=readOutputLines();
		//		requestRead();
		//	}
		//	catch(Exception e){}
		//	Thread.yield();
		//}
		
	}

	/** 
	 * Read the output lines. 
	 *
	 * return the bytes reads in from the device.
	 */
	public synchronized byte[] readOutputLines() throws IOException{
		byte bytes[]=new byte[2];
		
		bytes[0]=serialIO.read();
		bytes[1]=serialIO.read();
		
		return bytes;
	}
	
	/**
	 * Request a read from the output lines.
	 */
	public synchronized void requestRead() throws IOException {
		serialIO.write("!0RD");
	}

	/**
	 * Define the IO lines in the system.
	 *
	 * 1 mean output, 0 means input.
	 * param msb the most significant bit.
	 * param lsb the least significant bit.
	 */
	public synchronized void defineLines(byte msb, byte lsb) throws IOException{
		serialIO.write("!0SD"+(char)msb+(char)lsb);

		byte data[]=new byte[2];
		
	}

	/**
	 * Define the IO lines inital power on state.
	 *
	 * param msb the most significant bit.
	 * param lsb the least significant bit.
	 */
	public synchronized void setPowerupStates(byte msb, byte lsb) throws IOException{
		serialIO.write("!0SS"+(char)msb+(char)lsb);
	}

	/**
	 * Read the configuration of the device.
	 *
	 * return The bytes in the configuration. IO Definitions followed by Powerup states.
	 */
	public synchronized byte[] readConfiguration() throws IOException{
		byte bytes[]=new byte[4];
		serialIO.write("!0RC");
		
		bytes[0]=serialIO.read();
		bytes[1]=serialIO.read();
		bytes[2]=serialIO.read();
		bytes[3]=serialIO.read();
		
		
		
		return bytes;
	}
	
	/**
	 * Clean out the data, so that all the devices are turned off.
	 */
	public synchronized void turnOff() {
		/* If the system does not have this device. */
		if(!deviceConnected)
			return;

		flushOutputData();
		try{
		
			requestRead();
			
			byte data[]=new byte[2];
			data[0]=-1;
			data[1]=-1;
			requestRead();
	 		
			System.err.println("Waiting for system to turn off the pinball machine's parts."); 
			resetOutputLines(ALL,ALL);			
			/* Wait until the commands are sent to the lsb of the data unit. */
			while(data[1]!=0) {
				/* Send the command again. */
				resetOutputLines(ALL,ALL);
				try{
					data=readOutputLines();
					requestRead();
				}
				catch(Exception e){requestRead();}
				Thread.yield();
			}
		}
		catch(Exception e){System.err.println("turnOff:"+e);}
	}


	/**
	 * Used to test the device. 
	 */	
	public static void main(String args[]) {
		new RS232SDD16(testPort).test();
		//new RS232SDD16(testPort).readTest();
	}
	
	/**
	 * Clear any data waiting to be read. 
	 */
	public synchronized void flushInputData() {
		serialIO.flushInputData();
	}

	/**
	 * Clear any commands waiting to go out. 
	 */
	public synchronized void flushOutputData() {
		serialIO.flushOutputData();
	}

	/**
	 * Test the reading of the port. 
	 */
	private void readTest(){
		byte data[];
		while(true){
			try {
				requestRead();
				data=readOutputLines();
				System.err.println("1 "+data[0]+" "+data[1]);
			}
			
			catch(Exception e){System.err.println(e);} 
		}
	}

	/**
	 * Test's the port.
	 */
	private void test() {
		byte data[];
		
		BitSet left=new BitSet(WORDSIZE),right=new BitSet(WORDSIZE),both=new BitSet(WORDSIZE);
		
		left.set(1);
		right.set(0);
		both.set(0);
		both.set(1);
		
		/* Have the flippers constanting flipping in a pattern to test them. */
		while(true) {
			try {
				setOutputLines(ZERO,left);
				Thread.sleep(1000);
				setOutputLines(ZERO,right);
				Thread.sleep(1000);
				resetOutputLines(ZERO,left);
				Thread.sleep(1000);
				resetOutputLines(ZERO,right);
				Thread.sleep(1000);
			
				requestRead();
				data=readOutputLines();
				System.err.println("1 "+data[0]+" "+data[1]);


				setOutputLines(ZERO,both);
				setOutputLines(ZERO,both);
				Thread.sleep(1000);
				resetOutputLines(ZERO,both);
				resetOutputLines(ZERO,both);
				Thread.sleep(1000);
				
				data=readOutputLines();
				System.err.println("2 "+data[0]+" "+data[1]);

				setOutputLines(ZERO,right);
				Thread.sleep(1000);
				setOutputLines(ZERO,left);
				Thread.sleep(1000);
				resetOutputLines(ZERO,right);
				Thread.sleep(1000);
				resetOutputLines(ZERO,left);
				Thread.sleep(1000);
				
				data=readOutputLines();
				System.err.println("3 "+data[0]+" "+data[1]);

			}
		
			catch(Exception e){System.err.println(e);} 			
		}	

	}

	/**
	 * Generate random characters. 
	 */
	private byte random() {
		byte c = (byte)(Math.random()*25);
		c+=65;
		return c;
	}
} 
