package pinball.io;

import java.io.*;
import java.util.*;
import pinball.*;

/**
 * Reads and writes data to the serial port. 
 */
public class SerialPort extends Thread{
	/** The output to the port. */
	private DataOutputStream serialPortOut;
	
	/** The input from the port. */
	private DataInputStream serialPortIn;	
	
	/** The area where serial data is stored. */
	private SerialIO serialIO;
	
	/**
	 *
	 *
	 * @param port the port of the serial port.
	 * @param io the location of to store and get the data for the port.
	 */
	public SerialPort(String port, SerialIO io) {

		/* Make serial port. */
		try {serialPortOut =  new DataOutputStream(new FileOutputStream(port));}
		catch (IOException e){
			System.err.println("Serial Port Error "+e);
			System.exit(1);
		}
		
		try {
	
			serialPortIn = new DataInputStream(new FileInputStream(port));
		}
		catch (Exception e){
			System.err.println("Serial Port Error "+e);
			System.exit(1);
		}		

		/* Where to store or get data from. */
		serialIO=io;
		
		System.err.println("Serial Port Made: "+port);		
	}

	/**
	 * Wake the thread up, data is available.
	 */
	public synchronized void wakeup() {
		try{notify();}
		catch(Exception e){System.err.println(e);}
	}

	/**
	 * The thread is waiting for data to become available. 
	 */
	private synchronized void pause() {
		try{wait();}
		catch(Exception e){System.err.println(e);}
	}
	
	/** 
	 * Keep doing IO on the serial port.
	 */
	public void run() {
		while(true) {
		
			pause();
			

			/* Read the input from the port. */
			if(dataAvailable()) {
				try{
					serialIO.setReadData(read());
				}				
				catch(Exception e) {
					System.err.println(e);
				}
			}
		
						
			/* While there are commands to be processed write them
			   to the port. Also check for and available data that
			   may have become available because of a read request.
			 */
			while(serialIO.outputAvailable()) {
				/* Write out the command. */
				try{
					write(serialIO.outputData());
				}
				catch(Exception e) {	
					System.err.println(e);
				}
				
				/* Read the input from the port. */
				if(dataAvailable()) {
					try{
						serialIO.setReadData(read());
					}				
					catch(Exception e) {
						System.err.println(e);
					}
				}
				
				Thread.yield();
			}											
		}
		
	}
	
	/**
	 * Checks if data is available in the port. 
	 *
	 * return the status of data.
	 */
	public synchronized boolean dataAvailable() {
		try{
			if(serialPortIn.available()>0)
				return true;
			else return false;				
		}
		catch(Exception e){System.err.println(e);return false;}
		
		
	}
	
	
	/**
	 * Read from the port.
	 *
	 * return the byte read from the port.
	 */
	public synchronized byte read() throws IOException{
		byte b=-1;
		if(dataAvailable())
			b=(byte) serialPortIn.read();
		return b;
	}

	/**
	 * Write the data to the port.
	 *
	 * param data the data to write.
	 */
	public synchronized void write(Vector data) throws IOException{
		for(int i=0; i<data.size(); i++) {
			serialPortOut.writeBytes((String)data.elementAt(i));
			//System.err.println((String)data.elementAt(i));		
			serialPortOut.flush();
			try{Thread.sleep(5);}
			catch(Exception e){System.err.println(e);}
		}
		Thread.yield();
		
	}

}
