package pinball.player;

import pinball.*;
import pinball.buffer.*; 
import pinball.part.*;
import pinball.server.*;
import pinball.player.*;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;


/**
 * The way the strategy communicates with the pinball machine.
 * It does this in an indirect way, so a strategy can play over the network.
 * The stub also filters out the ball from the objects found.
 */
public class Stub extends Thread {
	/** The buffer that hold the information buffer. */
	protected Buffer buffer;

	/** The information buffer. */
	protected BufferInfo buffinfo;
	
	/** The flippers in the game. */
	protected Flipper left, right;	
	
	/** The statistics for the game. */
	protected Stats stats;
	
	/** Location to write game stats. */
	protected TextArea gameStats;
	
	
	/** X location of the ball. */
	protected	int x;
	
	/** Y location of the ball. */
	protected 	int y;
	
	/** Image x size. */
	protected	int xsize; 
	
	/** Image y size. */
	protected	int ysize;
	
	/** Left flipper x location. */
	protected int leftX=0;
	
	/** Left flipper y location. */
	protected int leftY=0;
	
	/** Right flipper x location. */	
	protected int rightX=0;
	
	/** Right flipper y location. */	
	protected int rightY=0;
	
	/** Area around flipper that is to be ignored. */
	protected int boarder=1;
	
	/* Flipper names. */
	
	/** Value of the left flipper. */
	private final int LEFT=0;
	
	/** Value of the right flipper. */
	private final int RIGHT=1;
	
	/** Where is the flipper from the top of the screen. */
	protected int flipperOffset=0;
	
	/* Used in finding the flippers. When was it flipped, and how long can it stay up. */

	/** Time when the flipper was flipped in testing. */
	private long startTime;	
	
	/** How long the flipper can be up in testing. */
	private long pause=1000;
		
	/** How long can the flipper stay up in the game. */
	protected int resetTime=GameOptions.resetTime();
				
	/** Last ball's x location. */
	protected int lastBallX=-1;
	
 	/** Last ball's y location. */
	protected int lastBallY=-1;
	
	/** Number of frames with only bad objects. */
	protected int badCount=0;

	/** Y location of begin flipper zone. */
	protected int flipperZone;
	
	/** Length of flipper zone. */
	protected int flipperZoneSize=GameOptions.flipperZoneSize();
	
	/** Did the flipper hit the ball. */
	protected boolean hitFlipper=false;
	
	/** Used in options to see if the they happened once. */
	protected boolean firstTime=true;
	
	/** Location from top of screen to be considered in play. */
	protected int fromTop=15;
	
	/** Location from bottom of flippers to be considered out of play. */
	protected int fromBottom=3;
	
	/** Anything moving below this be considered for flipping. */
	protected int flipperHeight=0;
	
	/** 
	 * Location from bottom of screen to be considered out of play. 
	 * Gets value from fromTop.
	 */
	protected int top=fromTop;
	
	/** 
	 * Location from bottom of screen to be considered out of play.
	 * Gets value from fromBottom. 
	 */
	protected int outZone=0;

	/**
	 * Location from the bottom at the screen to be considered out of play.
	 */
	protected int bottom;
	
	/** ID of strategy. */
	protected String id="Standard";
	
	/** Used to test the x location of the ball. x location + xoffset */	
	protected int xTest;
	
	/** Used to test the y location of the ball. y location + yoffset */	
	protected int yTest;
	
	/** Ball's distance from the last known location of the ball. */
	protected int distanceFromLastSeen=0;
	
	/** 
	 * Ball's distance from either the left or right flipper, depending on which
	 * half of the screen it's on.
	 */
	protected int distanceFromFlippers=0;

			
	/** The strategy to work on. */
	protected Strategy strategy;
 
	/** Are the flippers being tested. */
	protected boolean testingFlippers=true;
	
	/** The queue of waiting buffers. */
	protected StrategyQueue queue=new StrategyQueue(this);
	
	/** The plunger in the system. */
	protected Plunger plunger;
	
	/** Starts a new game on the pinball machine. */
	protected StartButton startButton;
	
	/** The ball return device. */
	protected BallReturn ballReturn;
	
	/** The camera in the system. */
	protected Camera camera;		
	
	/** Number of tries to launch the ball. */
	protected int triedLaunch=0;

	/** Was the left flipper last up. */
	private boolean leftFlipperLastState=false;

	/** Was the right flipper last up. */
	private boolean rightFlipperLastState=false;
	
	/** The greyscale values for the painting of the image.*/
	protected ColorModel model;

	/** Send frame. */
	protected boolean sendFrame=false;
	
	/** Used for creating an image. */
	protected Canvas canvas=new Canvas();

	/**
	 * The stub for the strategy.
	 *
	 * @param buffer The first buffer for the information buffers.
	 * @param gameStats the area on the screen to write stats info.
	 * @param strategy the strategy thread.
	 * @param camera the camera for the system.
	 */
	public Stub(Buffer buffer, TextArea gameStats, Strategy strategy, Camera camera) {
		setupImage();
		this.camera=camera;
		stats=new Stats(this);
		plunger=new Plunger(GameOptions.getDataIO());
		startButton=new StartButton();
		ballReturn=new BallReturn(GameOptions.getDataIO());
		GameOptions.setPlunger(plunger);
		this.buffer=buffer;
		this.gameStats=gameStats;
		this.strategy=strategy;
		buffinfo=buffer.info();

		
		left=new Flipper(GameOptions.getDataIO(),Flipper.LEFT);
		right=new Flipper(GameOptions.getDataIO(),Flipper.RIGHT);

		strategy.setQueue(queue);
	}
		
	/** Needed for when this class is extended. */
	public Stub(){}
	
	
	/**
	 * Setup the color model for the frame from the camera. Used for a strategy that wants to see the 
	 * image from the camera. 
	 */
	public void setupImage() {
		System.err.println("Setting up the image");
	 	/* Create the image's color table. (Grey scale)*/
        	byte grey[] = new byte[64];
	        for (int i = 0; i < 64; i++) {
        	   grey[i] = (byte) (i*4);
	        }
		
		/* Pure white. */
		grey[63]= (byte) 255;
	      	model = new IndexColorModel(6, 64, grey, grey, grey);
	}
	
	/**
	 * Name of the object.
	 */
	public String toString() {
		return "Stub";
	}

		
	/** 
	 * Get the loctions of the objects found and determine which is the ball and flip
	 * the flippers based on these finding.
	 */
	public void run() {
		/* Write the player id to the log file. */
		PinballLog.headerInfo("PLAYER="+id,this);
		
		/* Get the lock. */
		getLock();
		
		/* Initalize the strategy. */
		strategy.initalize();
		
		/* Setup the stub for the system. */
		buffinfo.setStub(this);
		
		/* Release the lock. */
		releaseLock();
		
		/* Next buffer. */
		nextBuffer();	

		locateFlippers();
		
		/* Return the ball to the plunger. */
		ballReturn.returnBall();		
		try{Thread.sleep(500);}
		catch(Exception e){}		
		ballReturn.reset();
				
		/* Check the frame to see if flipping is needed. */
		while(true) {

			/* Refind the flippers. */
			if(GameOptions.refindFlippers()) {
				GameOptions.locateFlippers(true);
				locateFlippers();
				GameOptions.refindFlippers(false);
				GameOptions.locateFlippers(false);
				/* Reinitalize the strategy. */
				strategy.initalize();
			}

			
			/* Get the lock and get the info. */
			getLock();
				
			/* If not skipping the buffer, process it. */
			if(!buffer.skipped())	{
				/* If the system is not paused, figure out what to do based on the blips found in the frame. */
				if(!GameOptions.pause()) {	
					//long timeStart=System.currentTimeMillis();
					
					process();
					
					//long totalTime=System.currentTimeMillis()-timeStart;
					//System.err.println("Stub's time "+totalTime);
					
				}
				else if(GameOptions.pause()) {
					skipBuffer();
				}
		       							 
			}
			
			/* Skipped frame. */
			else if(buffer.skipped()) {
				skipBuffer();
			}
			
	       		/* Go to the next buffer. */
			nextBuffer();
			
			/* Skip frames. */
			//skip();
				
			/* Let another thread go. */
			Thread.yield();
		}
	}


	/** 
	 * Skip the buffer.
	 */
	private void skipBuffer() {
//		System.err.println("Stub skipped a buffer "+buffer.id());
		releaseLock();
	}


	/** 
	 * Locate the flippers in the game.
	 */	
	public void locateFlippers() {
		if(GameOptions.locateFlippers()) {
			/* Find the flippers. */
			findFlippers(GameOptions.findFlippers());		
		}		

		else readFlippersFile();
		
		if(GameOptions.writeFlippersFile())
			writeFlippersFile();

		/* Write out the flipper states. */			
		PinballLog.headerInfo("leftFlipperX="+left.x()+" leftFlipperY="+left.y(),this);
		PinballLog.headerInfo("rightFlipperX="+right.x()+" rightFlipperY="+right.y(),this);
		
		/* Set the boarder for the flippers. */
		GameOptions.flipperSpace(boarder);
		
		/* Setup the stub info into the strategy. */
		strategy.getStubInfo();
		
		GameOptions.foundFlippers(true);	
	}
			
	/**
	 * Process the blips found. Find the ball and decide on wheither to flip or not.
	 */	 
	public void process() {		
		/* Set the size of the image. */
		this.xsize=width();
		this.ysize=height();
	
		/* The location of the ball. */			
		x=blipx();
		y=blipy();
		
		
		/* Write the blips found to the log file. */
		storeBlips();			
		
		/* Filter out the ball from other objects found. */
		filter(false);	
		
		/* Write the location of the ball to the log file. */
		storeBall();
			
		x=blipx();
		y=blipy();
			
		/* Save the last possition of the ball. */
	        if(x!=-1 && y!=-1) {
			lastBallX=x;
			lastBallY=y;   
		}							
				
		/* Get the statistics for the ball. */
		statistics();			
		
		addBuffer(buffer);
		
		strategy.bufferAvailable();			
		
		/* Every 100 frames display the time statisics. */
		if(number()%100==0) {
			PinballLog.writeln(number(),time().toString(),1,this);
			System.err.println("\n"+time().toString()+"\n");
		}
	}
	
	/**
	 * Set the buffer used by the stub.
	 *
	 * @param buffer the buffer to be used.
	 */
	public void setBuffer(Buffer buffer) {
		this.buffer=buffer;
		buffinfo=buffer.info();
	}
	
	/**
	 * Get the recent buffer from the queue.
	 *
	 * @return the recent buffer.
	 */
	public Buffer getRecentBuffer() {
		Buffer b=queue.recent();
		buffinfo=b.info();
		return b;
	}
	
	/**
	 * Get the next buffer from the queue.
	 *
	 * @return the next buffer in the queue.
	 */
	public Buffer getNextBuffer() {
		Buffer b=queue.next();
		buffinfo=b.info();
		return b;
	}
	
	/**
	 * Adds a buffer to the queue.
	 *
	 * @param b the buffer to add.
	 */
	public void addBuffer(Buffer b) {
		queue.add(b);
	}
	
	/**
	 * Checks if the queue is empty.
	 *
	 * @return empty status.
	 */
	public boolean isQueueEmpty() {
		return queue.empty();
	}
	
	/**
	 * Get the frame of found objects. 
	 *
	 * @return the blips frame.
	 */
	public int [][] getBlipsFrame() {
		return buffinfo.getBlipsFrame();
	}
	
	/**
	 * Get the objects table. 
	 */
	public Vector getObjects() {
		return buffinfo.getObjects();
	}

	/**
	 * Set teh blips in the buffer. 
	 *
	 * @param b the blips found. 
	 */
	public void setObjects(Vector b) {
		buffinfo.setBlips(b);
	}

	/**
	 * Get the raw frame from the camera. 
	 */
	public byte[] getRawFrame() {
		return buffinfo.get();
	}

	/**
	 * Get frame grabbed number. 
	 *
	 * @return the number of this frame.
	 */	
	public int number() {
		return buffinfo.number();
	}

	/**
	 * Get buffer's id. 
	 *
	 * @return the id of this buffer.
	 */	
	public int id() {
		return buffinfo.id();
	}
		
	/**
	 * Get the ball's x location.
	 *
	 * @return ball's x location
	 */
	public int blipx() {
		return buffinfo.blip().x();
	}

	/**
	 * Get the ball's y location.
	 *
	 * @return ball's y location
	 */
	public int blipy() {
		return buffinfo.blip().y();
	}

	/**
	 * Set the ball's x location.
	 *
	 * @param x ball's x location
	 */
	public void blipx(int x) {
		buffinfo.blip().x(x);
	}

	/**
	 * Set the ball's y location.
	 *
	 * @param y ball's y location
	 */
	public void blipy(int y) {
		buffinfo.blip().y(y);
	}


	/** 
	 * The time stamp used in the threads. 
	 *
	 * @return the time stamp.
	 */
	public TimeStamp time() {
		return buffinfo.time();
	}
	

	
	/**
	 * Get the width of the frame. 
	 *
	 * @return width.
	 */
	public int width() {
		return buffinfo.width();
	}
	
	/**
	 * Get the height. 
	 *
	 * @return height.
	 */
	public int height() {
		return buffinfo.height();
	}	

	/**
 	 * Get the flipper boarder.
 	 * 
	 * @param return The size of the flipper boarder.
	 */
	public void boarder(int b) {
		boarder=b;
		GameOptions.flipperSpace(b);
	}
	
	/**
	 * Set the location of the offsets.
	 *
	 * @param x xoffset.
	 * @param y yoffset.
	 */
	public void offsets(int x, int y) {
		buffinfo.offsets(x,y);
	}	
	
	/**
	 * Get the resize x size.
	 *
	 * @return x resize.
	 */
	public int resizeX() {
		return buffinfo.resizeX();
	}

	/**
	 * Get the resize y size.
	 *
	 * @return y resize.
	 */	
	public int resizeY() {
		return buffinfo.resizeY();
	}

	/**
	 * Get the resize x offset.
	 *
	 * @return x resize offset.
	 */
	public int resizeXOffset() {
		return buffinfo.resizeXOffset();
	}

	/**
	 * Get the resize y offset.
	 *
	 * @return y resize offset.
	 */
	public int resizeYOffset() {
		return buffinfo.resizeYOffset();
	}


	/**
	 * Get the x offset.
	 *
	 * @return x offset.
	 */
	public int XOffset() {
		return buffinfo.XOffset();
	}
	
	/**
	 * Get the y offset.
	 *
	 * @return y offset.
	 */
	public int YOffset() {
		return buffinfo.YOffset();
	}


	/** 
	 * Set the resize flag and include resize information. 
	 *
	 * @param x x size.
	 * @param y y size.
	 * @param OX x offset.
	 * @param OY y offset.
	 */
	public void resize(int x, int y, int OX, int OY) {
		buffinfo.resize(x,y,OX,OY);
//		System.err.println("RESIZING IMAGE");
	}
	
	/**
	 * Get the resize flag. 
	 * 
	 * @return resize flag.
	 */
	public boolean resize() {
		return buffinfo.resize();
	}

	/**
	 * Set the condition that it was resized. 
	 */
	public void resized() {	
		buffinfo.resized();		
	}
	
	/** 
	 * Set the clearing flag. Will clear the frame on the display. Used
	 * to clear when the ball's location is draw near the edge of the display.
	 * Clears the cross hair that's left at the edge.
	 *
	 * @param c clearing status.
	 */
	public void clearing(boolean c) {
		buffinfo.clearing(c);
	}

	/** 
	 * Get the clearing flag. 
	 *
	 * @param clearing status.
	 */
	public boolean clearing() {
		return buffinfo.clearing();
	}

	/**
	 * Get the flipper offset. 
	 *
	 * Return flipper offset.
	 */
	public int flipperOffset() {
		return buffinfo.flipperOffset();
	}
	
	/**
	 * Set the flipper offset. 
	 *
	 * @param i flipper offset.
	 */
	public void flipperOffset(int i) {
		buffinfo.flipperOffset(i);
	}
	
	/**
	 * Set the image offset. 
	 *
	 * @param i the image offset.
	 */
	public void imageOffset(int i) {
		buffinfo.imageOffset(i);
	}
	
	/** 
	 * Get the image offset. 
	 *
	 * @return the image offset.
	 */
	public int imageOffset() {
		return buffinfo.imageOffset();
	}
	
	
	/**
	 * Get the flippers flipped condition. 
	 *
	 * @return the status of the flippers flipped status.
	 */
	public boolean flipperFlipped() {
		return buffinfo.flipperFlipped();
	}

	/**
	 * Set the top & bottom for the in play area. 
  	 * Top is the top of the screen to define anything over that is to be
  	 * in play.
  	 * Bottom is the end of the playing area. Anything under that is out of
  	 * play.
  	 * 
  	 * @param t top of playfield.
  	 * @param b bottom of playfield.
  	 */
	public void topBottom(int t, int b) {
		buffinfo.topBottom(t,b);
		top=t;
		bottom=b;
	}
	
	/**
	 * Override the outZone value.
	 *
	 * @param out the new outZone value.
	 */
	
	public void outZone(int out) {
		outZone=out;
	}
	
	/**
	 * The top of the playfield.
	 *
	 * @return top of playfield.
	 */
	public int top() {
		return top;
	}

	/**
	 * The bottom of the playfield.
	 *
	 * @return bottom of playfield.
	 */	
	public int bottom() {
		return bottom;
	}

	/**
	 * Go to the next information buffer.
	 */
	public void nextBuffer() {
		buffer=buffer.next();
	}
	
	/**
	 * Get the lock on the information buffer. 
	 */	
	public void getLock() {
		buffer.lock().strategy(buffer.id());
		buffinfo=buffer.info();
	}
	
	/**
	 * Release the lock on the information buffer. 
	 */
	public void releaseLock() {
//		System.err.println("Releasing lock");
		buffer.lock().strategied(buffer.id());	
	}


	/**
	 * Release the lock on the information buffer. 
	 *
	 * @param Buffer b The buffer to unlock.
	 */
	public void releaseLock(Buffer b) {
		releaseLock(b.id());	
	}
	
	/**
	 * Release the lock on the information buffer. 
	 *
	 * @param int b The buffer to unlock.
	 */
	public void releaseLock(int buff) {
//		System.err.println("Trying to release lock "+buff);		
		queue.release(buff);
	}
		
	public int getBuffNumber() {
 		return buffer.info().number();
	}
	
	/**
	 * Set the state of the left flipper.
	 *	 
	 * @param s Flip state. True for up, false for down.
	 */
	public void leftFlipperState(boolean s) {
		if(s==true)
			left.flip();
		else left.reset();
		if(!testingFlippers)
			strategy.flipperFlipped(true);
		else buffinfo.flipperFlipped(s);		
		
	}

	/**
	 * Set the state of the right flipper.
	 *	          
	 * @param s Flip state. True for up, false for down.
	 */
	public void rightFlipperState(boolean s) {
		if(s==true)
			right.flip();
		else right.reset();	
		if(!testingFlippers)
			strategy.flipperFlipped(true);
		else buffinfo.flipperFlipped(s);		
	}
	
	public void flipperFlipped(boolean s) {
		buffinfo.flipperFlipped(s);
	}
	
	/**
	 * Should the left flipper be reset. Has it been up long enough. 
	 *
	 * @param done Has the flipper been up at least this long?
	 *
	 * @return should the flipper be reset or not.
	 */
	public boolean leftFlipperReset(long reset) {
		return left.timeStamp().reset(reset);
	}
	
	/**
	 * Should the right flipper be reset. Has it been up long enough. 
	 *
	 * @param done Has the flipper been up at least this long?
	 *
 	 * @return should the flipper be reset or not.
	 */
	public boolean rightFlipperReset(long reset) {
		return right.timeStamp().reset(reset);		
	}

	/**
	 * Is the left flipper down.
	 *
	 * @return flipper's down status. 
	 */	
	public boolean isLeftUp() {
		return left.up();
	}

	/**
	 * Is the right flipper down.
	 *
	 * @return flipper's down status. 
	 */	
	public boolean isRightUp() {
		return right.up();
	}
	
	/**
	 * Is the left flipper down.
	 *
	 * @return flipper's down status. 
	 */
	public boolean isLeftDown() {
		return left.down();
	}
	
	/**
	 * Is the right flipper down.
	 *
	 * @return flipper's down status. 
	 */
	public boolean isRightDown() {
		return right.down();
	}
	
	/**
	 * Write the blips to the log file. 
	 */
	public void storeBlips() {
		/* The objects seen. */ 
		Vector blips=getObjects();
		
		
		if(blips==null)
			return;
		
		/* The blip working on. */
		Blip blip;
		
		/* Amount of blips found. 
		*/
		int amount=0;
		
		if(blips!=null)
			amount=blips.size();
		
		/* Write the blip's info to the log file. */
		PinballLog.writeln(this.number(),"objects="+amount,1,this);
		for(int t=0;t<blips.size();t++) {
			blip=(Blip)blips.elementAt(t);
			PinballLog.writeln(this.number(),"object="+(t+1)+" alias="+blip.alias()+" Xlocation="+(blip.xLocation()+this.XOffset())+" Ylocation="+(blip.yLocation()+this.YOffset())+" size="+blip.total()+" minX="+blip.minX()+" minY="+blip.minY()+" maxX="+blip.maxX()+" maxY="+blip.maxY(),1,this);
			
			/* For debugging perposes. */
			//System.err.println("OBJECT SIZE "+blip.total());
		}
	}



	/**
	 * The frame was skipped.
	 */
	public void skipFrame() {
			blipx(-1);
			blipy(-1);
	}

	/**
	 * Skip past some of the buffers, if there is more recent data in the system.
	 */
	public boolean skip() {
		int firstChecked=buffer.id();
		boolean skipped=false;
			
		/* Not skippable. */
		if(buffer.lock().status()!=buffer.lock().analyzed)
			return skipped;			
	
		
//		System.err.println("Trying to skip a frame.");
		/* Skip frames while the next buffer has a frame that's usable. */
		while((buffer.next().lock().status()==buffer.next().lock().analyzed) && (buffer.next().id()!=firstChecked)) {

			/* Let this buffer pass through, since a later one is available. */

			getLock();
			skipFrame();

			PinballLog.writeln(buffinfo.number(),"strategySkippingFrame="+buffer.info().number(),1,this);
			System.err.println("Strategy skipping frame #"+buffinfo.number());
			
			buffer.skipped(true);
			
			releaseLock();
			nextBuffer();

						
		 }
		 return skipped;
	}
	

	/**
	 * Save the ball's location to the log file. 
	 */
	public void storeBall() {
		/* The ball found. */
		Blip blip=buffinfo.blip();
		
		/* Ball is found. */
		if(blip.x()!=-1)
			PinballLog.writeln(this.number(),"BALL-LOCATION ballXlocation="+(blip.x()+this.XOffset())+" ballYlocation="+(blip.y()+this.YOffset()),1,this);
	}

	
	/**
	 * Filter out the ball from the other objects found. 
	 * 
	 * @param findFlippers Is the system finding the flippers.
	 */
	public void filter(boolean findFlippers) {
		/* The objects found. */
		Vector blips=getObjects();
		
		/* The object. */
		Blip thing;
		
		/* Object concidered to be the ball in the Vector of objects. */
		int ball=-1;
		
		/* Used in loops. */
		int i;		
		
		/* Closest object found. */
		int closest=99999;
		
		/* Number of bad objects. */
		int badObjects=0;
		
		/* Did the flipper change state. */
		boolean flipperChanged=false;
		
		/* Is the ball in play. */
		boolean ballInPlay=stats.ballInPlay();
		
		/* Middle of the flippers. */
		int middleFlippers=(leftX+rightX)/2;		

					
		/* Search the objects for the ball. */
		if(ball==-1) {
			for (i=0; i < blips.size(); i++) { 
				thing=(Blip)blips.elementAt(i);		
				
				//System.err.println("Size: "+thing.width()+" "+thing.height());	
				
				//System.err.println("Checking Object. ");
				
				x=thing.xLocation();
				y=thing.yLocation();								
				
				xTest=x+XOffset();
				yTest=y+YOffset();
				
				//System.err.println("Location: "+xTest+" "+yTest);	
				
				//System.err.println(xTest+","+yTest);
				
				/* Distance from left flipper. */
				if(x<=middleFlippers) {
					distanceFromFlippers=(int)Math.sqrt(Math.pow(leftX-xTest,2)+Math.pow(leftY-yTest,2));
					if(isLeftUp()!=leftFlipperLastState) {
						leftFlipperLastState=isLeftUp();
						flipperChanged=true;
					}
				}

				/* Distance from right flipper. */
				if(x>middleFlippers) {
					distanceFromFlippers=(int)Math.sqrt(Math.pow(rightX-xTest,2)+Math.pow(rightY-yTest,2));
					if(isRightUp()!=rightFlipperLastState) {
						rightFlipperLastState=isRightUp();
						flipperChanged=true;
					}
				}

				distanceFromLastSeen=(int)Math.sqrt(Math.pow(x-lastBallX,2)+Math.pow(y-lastBallY,2));
				PinballLog.writeln(number(),"object="+(i+1)+" distanceFromLastSeen="+distanceFromLastSeen,1,this);
				
				
				/* Find the ball. It must be the right size and proper distance from the flippers. */
				if(thing.width()<=GameOptions.ballMax() &&
				thing.height()<=GameOptions.ballMax() && distanceFromFlippers>boarder && !flipperChanged) {								
					PinballLog.writeln(number(),"object="+(i+1)+" distanceFromFlippers="+distanceFromFlippers,1,this);
					/* If ball is in play, check to see if
					ball moved too far. */
					if(ballInPlay && lastBallX!=-1
					&& lastBallY!=-1 && x!=-1 && y!=-1 && x!=lastBallX && y!=lastBallY) {					
					
						/* If it moved within the
						allowed amount it's the ball. */
						if(distanceFromLastSeen<GameOptions.moveThreshold()){
									
							if(distanceFromLastSeen<closest)
							{											
								ball=i;
								closest=distanceFromLastSeen;
								badCount=0;
							}
						}
							
						/* Ball moved too much. */
						else {
								PinballLog.writeln(number(),"badObjectMovedTooMuch="+(i+1),1,this);
								badObjects++;
						}				
						
					}
					
					/* If ball is not in play, just check for any object on the screen. */			
					else if(!ballInPlay){					
						ball=i;
						closest=distanceFromLastSeen;
						PinballLog.writeln(number(),"ballAtTopOfScreen="+(i+1),1,this);						
					}
					
					/* Ball must have been lost, take any object. */
					else if(ballInPlay && lastBallX==-1 && lastBallY==-1){									
						ball=i;
						closest=distanceFromLastSeen;
						PinballLog.writeln(number(),"thinksBallMightBe="+(i+1),1,this);						
					}
					
				}
				
				/* Object is ignored because it's not the right size or too close to the flippers. */
				else if(ballInPlay) {
					if(distanceFromFlippers<=boarder){					
						PinballLog.writeln(number(),"ignoredObjectTooCloseToFlippers="+(i+1),1,this);
						badObjects++;
					}												
				
					else if(thing.total()<GameOptions.ballMin()){					
						PinballLog.writeln(number(),"ignoredObjectTooSmall="+(i+1),1,this);
						badObjects++;
													
					}
				
					else if(thing.total()>GameOptions.ballMax()){					
						PinballLog.writeln(number(),"ignoredObjectTooBig="+(i+1),1,this);
						badObjects++;
												
					}
				}
				
				else if (!ballInPlay){
					if(!findFlippers) {
						PinballLog.writeln(number(),"ignoredObjectWrongPlaceWrongTime="+(i+1),1,this);
						badObjects++;
					}
	
					/* Checking flippers, take smallest object. */
					else if (findFlippers) {
						ball=i;
						closest=distanceFromLastSeen;
						badCount=0;
					}
				}
				
			}
		}

		//System.err.println(badObjects+" "+blips.size());	
		
		/* Too many bad balls. Ball was lost. */
		if(blips.size()>0 && badObjects==blips.size()) {
			badCount++;
			PinballLog.writeln(number(),"frameBad="+badCount,1,this);
			/* Ball may be lost. Number of frames of bad objects is too much.*/
			if(badCount==5) {
				lastBallX=-1;
				lastBallY=-1;
				badCount=0;
				PinballLog.writeln(number(),"ballLost",1,this);
			}
		}
			
		/* Ball is found. */
		if(ball>-1) {
			
			thing=(Blip)blips.elementAt(ball);
			x=thing.totalX() / thing.total();
			y=thing.totalY() / thing.total();
			blipx(x);
			blipy(y);
			//System.err.println("Object's Size: "+thing.total());
		}
		
		/* No ball. */
		else {
			x=-1;
			y=-1;
			blipx(-1);
			blipy(-1);			
		}
	}
	
	/**
	 * Check the list of statistic events and see what to log. 
	 */
	public void statistics() {
		
		int ylocation=y+YOffset();
		
		boolean ballInPlay=stats.ballInPlay();
		
		/**********************************/
		/***** Log the flipper states. ****/
		/**********************************/			
		if(isLeftUp())
			PinballLog.writeln(number(),"leftFlipperUp",1,this);
		if(isRightUp())
			PinballLog.writeln(number(),"rightFlipperUp",1,this);
	
		/* Attempt to launch the ball. */
		if(!ballInPlay && stats.ballLaunchable() || GameOptions.launchBall()) {
			launchBall();
		}
		 
		/* If ball is launched. */
		//if(plunger.ballLaunched() && !ballInPlay && ylocation>0 && ylocation<top){			
		if(plunger.ballLaunched() && !ballInPlay){			
			stats.newBall();
			System.err.println("***BALL IN PLAY***");
			PinballLog.writeln(number(),"BALL_IN_PLAY",1,this);
		}				

		/* Ball was near the flipper. */
		if(ballInPlay && hitFlipper && ylocation!=-1) {
			/* See if the ball's in the flipper zone. If so, it was hit. */
			if(ylocation<(flipperZone+flipperZoneSize) && ylocation>flipperZone) {
				System.err.println("Ball was hit");
				PinballLog.writeln(number(),"BALL_HIT",1,this);
				stats.hitsInc();
				hitFlipper=false;
			}
		}

		/* If ball fell down. */
		//if((ballInPlay && ballReturn.ballInDrain()) || (GameOptions.ballOutOfPlay())) {
		if((ballInPlay && ballReturn.ballInDrain()) || (GameOptions.ballOutOfPlay())) {
			stats.ballDone();	

						
			triedLaunch=0;
			
			hitFlipper=false;
			System.err.println("***BALL IS GONE***");
			PinballLog.writeln(number(),"BALL_OUT_OF_PLAY",1,this);
			
			if(!GameOptions.ballOutOfPlay())
				PinballLog.writeln(number(),"type=natural",1,this);
			else if(GameOptions.ballOutOfPlay())
				PinballLog.writeln(number(),"type=forced",1,this);

			/* Write to screen. */
			Vector gameStatsInfo=new Vector();
			long time=stats.ballTime();
			float avgTime;
			gameStatsInfo.addElement("BALL #"+stats.balls());						
			gameStatsInfo.addElement("\nBall was in play for "+time+".");
			
			gameStatsInfo.addElement("\nBall was hit "+stats.hits()+" times.");
						
			avgTime=stats.avgBallTime();
			gameStatsInfo.addElement("\nAverage ball play time is "+avgTime+".");
			gameStatsInfo.addElement("\nAverage number of hits per ball is "+stats.avgHits());


			/* Write to statisics log file. */
			PinballLog.writeln(number(),"STATS totalBalls="+stats.balls()+" totalTime="+stats.totalTime(),1,this);
			time=stats.ballTime();
			PinballLog.writeln(number(),"STATS Ball="+stats.balls()+" inPlay="+time+" Hits="+stats.hits(),1,this);						
			avgTime=stats.avgBallTime();
			PinballLog.writeln(number(),"STATS avgInPlay="+avgTime+" avgHits="+stats.avgHits(),1,this);
			
			stats.statsDone();
			
			PinballLog.newBall();
			camera.nextBall();
			
			/* Write out the game stats to the screen. */
			String string;
			string=(String)gameStatsInfo.elementAt(0);
			gameStats.setText(string);
			for(int i=1;i<gameStatsInfo.size();i++) {
				string=(String)gameStatsInfo.elementAt(i);
				gameStats.append(string);
			}

			
			/* This clears up the ball's excess that's drawn near the
			   edges of the screen.
			*/
			clearing(true);

		}
		
		/* Ball is near the flipper. */
		if(ballInPlay && ylocation>(flipperZone+flipperZoneSize)) {
			hitFlipper=true;
			PinballLog.writeln(number(),"ball_near_flipper",1,this);
		}
		
		/* Ball's moving. 
		if(ballInPlay && x!=-1 && y!=-1)
			stats.ballAlive();
		*/
		
		/* Ball is not found. 
		if(ballInPlay && x==-1 && y==-1) 
			stats.ballDisappeared();
		*/
		
			
		
//		System.err.println(ylocation+" "+flipperZone+" "+flipperZoneSize);
	}
			
	
	/**
	 * Search for the flipper the number of times given. 
	 *
	 * @param quantity number of times to look for the flippers. 
	 */
	public void findFlippers(int quantity) {
		int ignoreAmount=3;
		
		/* Current flipper test number. 
		   Ignore the first few times the flippers are checked. 
		*/
		int at=-ignoreAmount; 
		
		/* X flipper test values. */	
		int []x=new int[quantity];
		
		/* Y flipper test values. */
		int []y=new int[quantity];
		
		/* Still testing flippers? */
		boolean findingFlippers=true;
		
		/* Used to sum up the values. */
		int total=0;
		
		/* Which flipper is being tested. */
		int testing=LEFT;
		
		/* Center of the change of the image. Used to find the center of the flipper. */
		Location center;
		
		System.err.println("Testing Flippers.");
		startTime=System.currentTimeMillis();
		
		/* Find the flipeprs. */
		while(findingFlippers) {
			getLock();		
			
			/* Write the blips found to the log file. */
			storeBlips();
	
			filter(true);	
	
			/* See if it's time to reset the left flipper. */	
			if(leftFlipperReset(pause) && isLeftUp()) {
				leftFlipperState(false);
				startTime=System.currentTimeMillis();
			}
			
			/* See if it's time to reset the right flipper. */				
			else if(rightFlipperReset(pause) && isRightUp()) {
				rightFlipperState(false);
				startTime=System.currentTimeMillis();	
			}

			/* Test the right flipper. */
			else if(testing==RIGHT && isRightDown() &&
			System.currentTimeMillis()-startTime>=pause) {
				rightFlipperState(true);
				startTime=System.currentTimeMillis();
								
				try {
        		       		Thread.sleep(1000);
	      			} 
		       		catch (InterruptedException e)
		       		{System.err.println(e);}
				
			}
			/* Test the left flipper. */
			else if(testing==LEFT  && isLeftDown() &&
			System.currentTimeMillis()-startTime>=pause) {				
				leftFlipperState(true);
				startTime=System.currentTimeMillis();
				
				try {
        		       		Thread.sleep(1000);
	      			}  
		       		catch (InterruptedException e)
		       		{System.err.println(e);}
			
			}
				
			center=getCenterChange();
			System.err.println("Buffer: "+getBuffNumber()+" "+center.x()+","+center.y());

			/* Store flipper locations. -1 means not found. */
			if(center.x()!=-1 && center.y()!=-1) {
				if(at>-1) {													
					x[at]=center.x();
					y[at]=center.y();
					blipx(center.x());
					blipy(center.y());
//					System.err.println("Movement "+x[at]+","+y[at]);
					PinballLog.writeln(number(),"FLIPPER FlipperX="+x[at]+" FlipperY="+y[at],1,this);
				}				
				at++;
			}

			/* Done with the left flipper. */			
			if(at==quantity && testing==LEFT) {
			
				/* Set up to test right flipper. */				
				testing=RIGHT;
				at=-2;
				total=0;
				
				leftFlipperState(false);
				
				/* Average the x. */
				for(int i=0;i<quantity;i++)
					total+=x[i];
				left.x(total/quantity);
				
				total=0;
				
				/* Average the y. */
				for(int i=0;i<quantity;i++)
					total+=y[i];
				left.y(total/quantity);
				
				System.err.println("Left flipper at: "+left.x()+","+left.y());				
				
				left.x(left.x());
				left.y(left.y());

				leftX=left.x();
				leftY=left.y();
				
				try {
        		       		Thread.sleep(1000);
	      			} 
		       		catch (InterruptedException e)
		       		{System.err.println(e);}
				
				/*Keep waiting until the left flipper stopped moving. */
				while(center.x()!=-1 && center.y()!=-1) {
					System.err.println("Ignored "+getBuffNumber()+" "+center.x()+","+center.y());
					releaseLock();
					nextBuffer();
					getLock();
					center=getCenterChange();
					time().reset();																													
					Thread.yield();
					try {
         			       		Thread.sleep(1000);
		      			}
					catch (InterruptedException e)
			       		{System.err.println(e);}
				}
					
			}
			
			
			/* Done with the right flipper. */
			else if(at==quantity && testing==RIGHT) {
				findingFlippers=false;
				
				total=0;
				
				/* Average the x. */
				for(int i=0;i<quantity;i++)
					total+=x[i];
				right.x(total/quantity);
				
				total=0;
	
				/* Error occured. Right x is less than Left x.
				*/			
				if(right.x()<left.x()) {
					System.err.println("Refinding flippers.");
					PinballLog.headerInfo("flipperError=Refinding_Flippers",this);
					testing=LEFT;
					at=-ignoreAmount;
					total=0;
					findingFlippers=true;
				}
					
				
				/* Average the y. */
				for(int i=0;i<quantity;i++)
					total+=y[i];
				right.y(total/quantity);
				System.err.println("Right flipper at: "+right.x()+","+right.y());
				
				rightFlipperState(false);

				right.x(right.x());
				right.y(right.y());
								
				if(right.y()<left.y())
					flipperOffset(right.y());
				else flipperOffset(left.y());
				
				rightX=right.x();
				rightY=right.y();
				
			}
			
			
			time().reset();
			releaseLock();
											
			nextBuffer();
		
			Thread.yield();
		}
		
		/* Find the higher flipper and then set the hight to start to draw
		   the flipper zone to test for flipper hits. 
		*/
		if(leftY>rightY) {
			flipperZone=leftY-GameOptions.flipperThreshold();			
			outZone=rightY+fromBottom;
		}
		else {
			flipperZone=rightY-GameOptions.flipperThreshold();
			outZone=leftY+fromBottom;
		}
		
		/* Set the flipper zone area. */
		GameOptions.flipperZone(flipperZone);

		System.err.println("Done testing.");
		
	}	
	
	/** 
	 * Check if ball is in play.
	 *
	 * @return ball's in play status.
	 */
	public boolean statsBallInPlay() {
		return stats.ballInPlay();
	}
	
	/** 
	 * Force ball in play
	 */
	public void ballIsInPlay() {
		stats.ballInPlay(true);
	}

	
	/** 
	 * X location of left flipper.
	 *
	 * @return left flipper's x location.
	 */
	public int leftX() {
		return leftX;
	}

	/** 
	 * Y location of left flipper.
	 *
	 * @return left flipper's y location.
	 */
	public int leftY() {
		return leftY;
	}
	
	/** 
	 * X location of right flipper.
	 *
	 * @return right flipper's x location.
	 */
	public int rightX() {
		return rightX;
	}

	/** 
	 * Y location of right flipper.
	 *
	 * @return right flipper's y location.
	 */
	public int rightY() {
		return rightY;
	}
	
	/**
	 * Get the flipper's zone.
	 *
	 * @return flipper zone.
	 */
	public int flipperZone() {
		return flipperZone;
	}

	/**
	 * Get the flipper's outzone.
	 *
	 * @return flipper outzone.
	 */	
	public int outZone() {
		return outZone;
	}

	/**
	 * Write to the log.
	 * @param frame frame number.
	 * @param info String to write.	 
	 * @param priority priority level.
 	 * @param object caller of logit.
	 */	
	public void logit(int frame, String info, int priority, Object object) {
		PinballLog.writeln(frame,info,priority,object);
	}
	
	/**
	 * Write to the header of the log.
	 *
	 * @param s String to write.
	 */		
	public void headerInfo(String s) {
		PinballLog.headerInfo(s,this);
	}

	/**
	 * For the flippers, to find the center of the change of the image.
	 * @return The location of the center of the image.
	 */
	public Location getCenterChange() {
		System.err.println("Finding center of image...");
		
		/* The objects found. */
		Vector blips=getObjects();
		
		Blip thing;
		
		Location center=new Location();
		
		int xTotal=0;
		int yTotal=0;
		
		for (int i=0; i < blips.size(); i++) {
			thing=(Blip)blips.elementAt(i);
				
			x=thing.totalX() / thing.total();
			y=thing.totalY() / thing.total();
				
			xTest=x+XOffset();
			yTest=y+YOffset();
			
			xTotal+=xTest;
			yTotal+=yTest;
		}
		
		if(blips.size()>0) {
			center.x(xTotal/blips.size());
			center.y(yTotal/blips.size());
		}
		
		return center;
	}
		
	/**
	 * Read in the flipper's file to see where the flippers were previously located.
	 */
	public void readFlippersFile() {
		DataInputStream flippersFile=GameOptions.getFlippersFileRead();
		try {			
			/* Left flipper X & Y. */
			left.x(leftX=flippersFile.readInt());
			left.y(leftY=flippersFile.readInt());

			/* Right flipper X & Y. */
			right.x(rightX=flippersFile.readInt());
			right.y(rightY=flippersFile.readInt());
						
			flippersFile.close();
		}
		catch (Exception e) {System.err.println(e);}
		
		/* Find the higher flipper and then set the hight to start to draw
		   the flipper zone to test for flipper hits. 
		*/
		if(left.y()>right.y()) {
			flipperZone=left.y()-GameOptions.flipperThreshold();			
			outZone=right.y()+fromBottom;
		}
		else {
			flipperZone=right.y()-GameOptions.flipperThreshold();
			outZone=left.y()+fromBottom;
		}
		
		/* Set the flipper zone area. */
		GameOptions.flipperZone(flipperZone);

		System.err.println("Done testing.");
		
		
	}
	
	/**
 	 * Write the flippers file, to store the locations of the flippers.
	*/
	public void writeFlippersFile() {
		System.err.println("Saving Flipper's State");
		DataOutputStream flippersFile=GameOptions.getFlippersFileWrite();
		try {			
			/* Left flipper X & Y. */
			flippersFile.writeInt(left.x());
			flippersFile.writeInt(left.y());

			/* Right flipper X & Y. */
			flippersFile.writeInt(right.x());
			flippersFile.writeInt(right.y());
			flippersFile.flush();
			flippersFile.close();
		}
		catch (Exception e) {System.err.println(e);}
		
	}
	
	/**
	 * Allows the strategy to draw to the screen any lines or other graphics
	 * for the user to see how the strategy operates.
	 * This also draws the flipper's locations.
	 *
	 * @param g the graphics area to draw to.
	 * @param canvasWidth the width of the canvas.
	 * @param canvasHeight the height of the canvas.
	 */	
	public void strategyDraw(Graphics g, int wSize, int hSize) {
		/* Draw the locations of the flippers. If they are found. */
		if(left.y()>-1 && right.y()>-1) {
			g.setColor(Color.red);
			int space=GameOptions.flipperSpace();
			if (space<2)
				space=2;
			int halfSpace=space/2;
			g.fillOval(left.x()-halfSpace,left.y()-halfSpace,GameOptions.flipperSpace(),space);
			g.fillOval(right.x()-halfSpace,right.y()-halfSpace,GameOptions.flipperSpace(),space);
		}	

		strategy.draw(g,wSize,hSize);
	}
	
	/**
	 * Have the system launch the ball. 
	 */
	public void launchBall() {
		if(!GameOptions.launchBall())
			PinballLog.writeln(number(),"type=natural",1,this);
		else if(GameOptions.launchBall())
			PinballLog.writeln(number(),"type=forced",1,this);			

		/* Return the ball to the plunger. */
		ballReturn.returnBall();	

		try{Thread.sleep(2000);}
		catch(Exception e){}
		
		
		plunger.launchBall();
		triedLaunch++;
		if(triedLaunch==3) {
			startButton.press();
			triedLaunch=0;
		}			
	}
	
	/**
	 * Get a frame from the camera. 
	 */
	public Image getImage(){
		/* If there is a frame. */
		try{
			if(buffinfo!=null && buffinfo.get()!=null) {
				/* Make the image.*/
			        MemoryImageSource mis = 
        				new MemoryImageSource(buffinfo.width(),buffinfo.height(),
  			   		   		model, buffinfo.get(), 0, buffinfo.width());

		        	/* The image from the camera. */
				Image myImage = canvas.createImage(mis);
					
				/*System.err.println(new String(buffinfo.get()));
				
				byte data[]=buffinfo.get();
				for(int i=0;i<data.length;i++)
					System.err.print(" "+data[i]);
				*/
				return myImage;	
			}
			else {
				return null;
			}
		}
		catch (Exception e){return null;}
		
		
	}
	

	/**
	 * Get a frame from the camera. 
	 */
	 public void requestFrame() {
	}
	
	/**
	 * Set if the system should send a frame from the camera over the network. 
	 */
	 public void sendFrame(boolean b) {
	 	sendFrame=b;
	}	
	
	/**
	 * Should the system send a frame from the camera over the network. 
	 */
	 public boolean sendFrame() {
		return sendFrame;
	}	
	
	/** 
	 * Release all the buffers from the queue. 
	 */
	public void releaseAllBuffers() {
		queue.releaseAll();
	}
}
