package pinball.player.adjustable;

import pinball.*;
import pinball.buffer.*; 
import pinball.part.*;
import pinball.server.*;
import pinball.player.*;
import pinball.player.standard.*;
import pinball.player.network.*;
import pinball.player.remote.*;

import java.awt.*;

/**
 * Player which can have the flipper options adjusted. 
 *
 * An additional command line argument is available for this player: 
 * -Pfile is used to store the locations of the polygons for the flippers.
 * <P>
 * While the system is running, press 1 or 2 to select the left or right player.
 * Then click on the window to define the outline of the zone for that flipper.
 * If you want to delete the outline, press d. If you want to save the outlines,
 * press s. The file specified by the command line argument will automatically,
 * load previous outlines, if they exist, and will save to that file name, if
 * save is chosen. 
 */
public class AdjustablePlayer extends StdPlayer{
	/** Id of class. */
	static { id="Adjustable"; }
	
	/** 
	 * Start up the program. Takes command line arguments.
   	 * Takes the following command line arguments:
	 * -s for server mode.
	 * -c for client mode.	
	 *
	 * New command line argument -Ppolygonfile for the locations 
	 * of the polygons for the flipper zones.
	 */
	public static void main(String args[]) {		
		StdPlayer player=new AdjustablePlayer();
		player.makeIt(player,args);
	} 
	
	
	
	/** 
	 * Get the strategy thread for the player. 
	 *
 	 * @param buffer the first buffer in the series of buffers for the system.
	 * @param gameStats the area to write the game stats to.
	 * @param camera the camera for the system.
 	 *
 	 * @return strategy thread.
	 */
	public Strategy localStrategy(Buffer buffer,TextArea gameStats,Camera camera) {
		return new AdjustableStrategy(buffer,gameStats,camera);
	}
	
	
	/** 
	 * Get the strategy thread for the player. 
	 *
 	 * @param buffer the first buffer in the series of buffers for the system.
	 * @param gameStats the area to write the game stats to.
 	 *
 	 * @return strategy thread.
	 */
	public Strategy clientStrategy(Buffer buffer,TextArea gameStats) {
		return new AdjustableRemoteStrategy(buffer,gameStats);
	}
	
	
	/** 
	 * Get the strategy thread for the player. 
	 *
 	 * @param buffer the first buffer in the series of buffers for the system.
	 * @param gameStats the area to write the game stats to.
	 * @param camera the camera for the system.
 	 *
 	 * @return strategy thread.
	 */
	public Strategy serverStrategy(Buffer buffer,TextArea gameStats,Camera camera) {
		return new NetworkPlayerStrategy(buffer,gameStats,camera);
	}	
	
	/** 
	 * Get the post thread for the player. 
	 *
 	 * @param buffer the first buffer in the series of buffers for the system.
 	 *
 	 * @return strategy thread.
	 */
	public Post post(Buffer buffer){
		return new AdjustablePost(buffer);
	}

	/**
	 * Make the information buffers. Return the first of the list.
	 *
	 * @param amount The amount of buffers to create.
	 * @param Qcam The camera.
	 *
	 * @return the first of the list.
	 */
	public Buffer makeBuffers(int amount, Camera Qcam) {
	        Buffer first=null; /* First buffer. */
		BufferInfo buffinfo=null; /* Buffer information. */
		Buffer buffer=null; /* The buffer being worked on. */
	        
	        /* Create the firstr buffer, with it's info. */
 		first=new AdjustableBuffer(1);
 		buffer=first;
		buffinfo=buffer.info();
		   
		System.err.println("Making "+amount+" buffers.");
		
		/* If less than 1 buffer. */
		if(amount<1) {
			System.err.println("Amount of frame buffer is less than 1!");
			System.exit(1);
		}
		
		/* Make the rest of the linked list of buffers. */
		for (int i=1;i<amount;i++) {
		
				/* Set up the information. */
				buffer.next(new AdjustableBuffer(i+1));
				buffinfo.cameraInfo(Qcam.frameSize(),Qcam.xSize(),Qcam.ySize());
				buffer=buffer.next();
				buffinfo=buffer.info();
		}
	
		/* Close the link. */				
		buffer.next(first);
		
		/* Set up it's information. */
		buffinfo.cameraInfo(Qcam.frameSize(),Qcam.xSize(),Qcam.ySize());
		
		
		/* Return the first buffer in the list. */
		return first;
	}


}
